from numpy import inf, nan
from sklearn.preprocessing import MinMaxScaler as Op

from lale.docstrings import set_docstrings
from lale.operators import make_operator


class _MinMaxScalerImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)


_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "inherited docstring for MinMaxScaler    Transforms features by scaling each feature to a given range.",
    "allOf": [
        {
            "type": "object",
            "required": ["feature_range", "copy"],
            "relevantToOptimizer": ["copy"],
            "additionalProperties": False,
            "properties": {
                "feature_range": {
                    "XXX TODO XXX": "tuple (min, max), default=(0, 1)",
                    "description": "Desired range of transformed data.",
                    "type": "array",
                    "laleType": "tuple",
                    "default": (0, 1),
                },
                "copy": {
                    "type": "boolean",
                    "default": True,
                    "description": "Set to False to perform inplace row normalization and avoid a copy (if the input is already a numpy array).",
                },
            },
        }
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Compute the minimum and maximum to be used for later scaling.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "The data used to compute the per-feature minimum and maximum used for later scaling along the features axis.",
        }
    },
}
_input_transform_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Scaling features of X according to feature_range.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "Input data that will be transformed.",
        }
    },
}
_output_transform_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Scaling features of X according to feature_range.",
    "laleType": "Any",
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://scikit-learn.org/0.20/modules/generated/sklearn.preprocessing.MinMaxScaler#sklearn-preprocessing-minmaxscaler",
    "import_from": "sklearn.preprocessing",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}
MinMaxScaler = make_operator(_MinMaxScalerImpl, _combined_schemas)

set_docstrings(MinMaxScaler)



import sklearn
from numpy import inf, nan
from sklearn.tree import DecisionTreeClassifier as Op

from lale.docstrings import set_docstrings
from lale.operators import make_operator


class _DecisionTreeClassifierImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)

    def predict_proba(self, X):
        return self._wrapped_model.predict_proba(X)


_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "inherited docstring for DecisionTreeClassifier    A decision tree classifier.",
    "allOf": [
        {
            "type": "object",
            "required": [
                "criterion",
                "splitter",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "min_weight_fraction_leaf",
                "max_features",
                "random_state",
                "max_leaf_nodes",
                "min_impurity_decrease",
                "min_impurity_split",
                "class_weight",
                "presort",
            ],
            "relevantToOptimizer": [
                "criterion",
                "splitter",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "max_features",
            ],
            "additionalProperties": False,
            "properties": {
                "criterion": {
                    "enum": ["entropy", "gini"],
                    "default": "gini",
                    "description": "The function to measure the quality of a split",
                },
                "splitter": {
                    "enum": ["random", "best"],
                    "default": "best",
                    "description": "The strategy used to choose the split at each node",
                },
                "max_depth": {
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimumForOptimizer": 3,
                            "maximumForOptimizer": 5,
                            "distribution": "uniform",
                        },
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "The maximum depth of the tree",
                },
                "min_samples_split": {
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimumForOptimizer": 2,
                            "maximumForOptimizer": 5,
                            "distribution": "uniform",
                        },
                        {
                            "type": "number",
                            "minimumForOptimizer": 2,
                            "maximumForOptimizer": 5,
                            "distribution": "uniform",
                        },
                    ],
                    "default": 2,
                    "description": "The minimum number of samples required to split an internal node:  - If int, then consider `min_samples_split` as the minimum number",
                },
                "min_samples_leaf": {
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimumForOptimizer": 1,
                            "maximumForOptimizer": 5,
                            "distribution": "uniform",
                        },
                        {
                            "type": "number",
                            "minimumForOptimizer": 1,
                            "maximumForOptimizer": 5,
                            "distribution": "uniform",
                        },
                    ],
                    "default": 1,
                    "description": "The minimum number of samples required to be at a leaf node",
                },
                "min_weight_fraction_leaf": {
                    "type": "number",
                    "default": 0.0,
                    "description": "The minimum weighted fraction of the sum total of weights (of all the input samples) required to be at a leaf node",
                },
                "max_features": {
                    "anyOf": [
                        {"type": "integer", "forOptimizer": False},
                        {
                            "type": "number",
                            "minimumForOptimizer": 0.01,
                            "maximumForOptimizer": 1.0,
                            "distribution": "uniform",
                        },
                        {"type": "string", "forOptimizer": False},
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "The number of features to consider when looking for the best split:      - If int, then consider `max_features` features at each split",
                },
                "random_state": {
                    "anyOf": [
                        {"type": "integer"},
                        {"laleType": "numpy.random.RandomState"},
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "If int, random_state is the seed used by the random number generator; If RandomState instance, random_state is the random number generator; If None, the random number generator is the RandomState instance used by `np.random`.",
                },
                "max_leaf_nodes": {
                    "anyOf": [{"type": "integer"}, {"enum": [None]}],
                    "default": None,
                    "description": "Grow a tree with ``max_leaf_nodes`` in best-first fashion",
                },
                "min_impurity_decrease": {
                    "type": "number",
                    "default": 0.0,
                    "description": "A node will be split if this split induces a decrease of the impurity greater than or equal to this value",
                },
                "min_impurity_split": {
                    "anyOf": [{"type": "number"}, {"enum": [None]}],
                    "default": None,
                    "description": "Threshold for early stopping in tree growth",
                },
                "class_weight": {
                    "XXX TODO XXX": 'dict, list of dicts, "balanced" or None, default=None',
                    "description": "Weights associated with classes in the form ``{class_label: weight}``",
                    "enum": ["balanced"],
                    "default": "balanced",
                },
                "presort": {
                    "type": "boolean",
                    "default": False,
                    "description": "Whether to presort the data to speed up the finding of best splits in fitting",
                },
            },
        },
        {
            "XXX TODO XXX": "Parameter: min_samples_leaf > only be considered if it leaves at least min_samples_leaf training samples in each of the left and right branches"
        },
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Build a decision tree classifier from the training set (X, y).",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
                    "XXX TODO XXX": "array-like or sparse matrix, shape = [n_samples, n_features]",
                },
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
            "description": "The training input samples",
        },
        "y": {
            "anyOf": [
                {"type": "array", "items": {"type": "number"}},
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
            "description": "The target values (class labels) as integers or strings.",
        },
        "sample_weight": {
            "anyOf": [{"type": "array", "items": {"type": "number"}}, {"enum": [None]}],
            "description": "Sample weights",
        },
        "check_input": {
            "type": "boolean",
            "default": True,
            "description": "Allow to bypass several input checking",
        },
        "X_idx_sorted": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
                {"enum": [None]},
            ],
            "default": None,
            "description": "The indexes of the sorted training input samples",
        },
    },
}
_input_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predict class or regression value for X.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
                    "XXX TODO XXX": "array-like or sparse matrix of shape = [n_samples, n_features]",
                },
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
            "description": "The input samples",
        },
        "check_input": {
            "type": "boolean",
            "default": True,
            "description": "Allow to bypass several input checking",
        },
    },
}
_output_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "The predicted classes, or the predict values.",
    "anyOf": [
        {"type": "array", "items": {"type": "number"}},
        {"type": "array", "items": {"type": "array", "items": {"type": "number"}}},
    ],
}
_input_predict_proba_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predict class probabilities of the input samples X.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
                    "XXX TODO XXX": "array-like or sparse matrix of shape = [n_samples, n_features]",
                },
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
            "description": "The input samples",
        },
        "check_input": {"type": "boolean", "description": "Run check_array on X."},
    },
}
_output_predict_proba_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "such arrays if n_outputs > 1",
    "laleType": "Any",
    "XXX TODO XXX": "array of shape = [n_samples, n_classes], or a list of n_outputs",
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://scikit-learn.org/0.20/modules/generated/sklearn.tree.DecisionTreeClassifier#sklearn-tree-decisiontreeclassifier",
    "import_from": "sklearn.tree",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "classifier"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
        "input_predict_proba": _input_predict_proba_schema,
        "output_predict_proba": _output_predict_proba_schema,
    },
}
DecisionTreeClassifier = make_operator(_DecisionTreeClassifierImpl, _combined_schemas)

if sklearn.__version__ >= "0.22":
    # old: https://scikit-learn.org/0.20/modules/generated/sklearn.tree.DecisionTreeClassifier.html
    # new: https://scikit-learn.org/0.22/modules/generated/sklearn.tree.DecisionTreeClassifier.html
    from lale.schemas import AnyOf, Bool, Enum, Float

    DecisionTreeClassifier = DecisionTreeClassifier.customize_schema(
        presort=AnyOf(
            types=[Bool(), Enum(["deprecated"])],
            desc="This parameter is deprecated and will be removed in v0.24.",
            default="deprecated",
        ),
        ccp_alpha=Float(
            desc="Complexity parameter used for Minimal Cost-Complexity Pruning. The subtree with the largest cost complexity that is smaller than ccp_alpha will be chosen. By default, no pruning is performed.",
            default=0.0,
            forOptimizer=False,
            minimum=0.0,
            maximumForOptimizer=0.1,
        ),
        set_as_available=True,
    )

if sklearn.__version__ >= "0.24":
    # old: https://scikit-learn.org/0.22/modules/generated/sklearn.tree.DecisionTreeClassifier.html
    # new: https://scikit-learn.org/0.24/modules/generated/sklearn.tree.DecisionTreeClassifier.html
    DecisionTreeClassifier = DecisionTreeClassifier.customize_schema(
        presort=None, set_as_available=True
    )

if sklearn.__version__ >= "1.0":
    # old: https://scikit-learn.org/0.24/modules/generated/sklearn.tree.DecisionTreeClassifier.html
    # new: https://scikit-learn.org/1.0/modules/generated/sklearn.tree.DecisionTreeClassifier.html
    DecisionTreeClassifier = DecisionTreeClassifier.customize_schema(
        min_impurity_split=None, set_as_available=True
    )

set_docstrings(DecisionTreeClassifier)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional, Union

from .schema_utils import JsonSchema


class SchemaRange(object):
    def __init__(
        self,
        minimum=None,
        maximum=None,
        exclusive_minimum=False,
        exclusive_maximum=False,
        is_integer: bool = False,
    ) -> None:
        self.minimum = minimum
        self.maximum = maximum
        self.exclusive_minimum = exclusive_minimum
        self.exclusive_maximum = exclusive_maximum
        self.is_integer = is_integer

    def __str__(self):
        res = ""
        if self.minimum is None:
            res += "(infty"
        else:
            if self.exclusive_minimum:
                res += "("
            else:
                res += "["
            res += str(self.minimum)
        res += ","
        if self.maximum is None:
            res += "infty)"
        else:
            res += str(self.maximum)
            if self.exclusive_maximum:
                res += ")"
            else:
                res += "]"
        return res

    @classmethod
    def point(cls, pt: Union[int, float]):
        return SchemaRange(
            minimum=pt,
            maximum=pt,
            exclusive_minimum=False,
            exclusive_maximum=False,
            is_integer=isinstance(pt, int),
        )

    @classmethod
    def fromSchema(cls, schema: Any) -> "SchemaRange":
        return SchemaRange(
            minimum=schema.get("minimum", None),
            maximum=schema.get("maximum", None),
            exclusive_minimum=schema.get("exclusiveMinimum", False),
            exclusive_maximum=schema.get("exclusiveMaximum", False),
            is_integer=schema.get("type", "number") == "integer",
        )

    @classmethod
    def fromSchemaForOptimizer(cls, schema: Any) -> "SchemaRange":
        s = cls.fromSchema(schema)

        minimum = schema.get("minimumForOptimizer", None)
        maximum = schema.get("maximumForOptimizer", None)
        exclusive_minimum = schema.get("exclusiveMinimumForOptimizer", False)
        exclusive_maximum = schema.get("exclusiveMaximumForOptimizer", False)
        is_integer = (
            schema.get("type", "numberForOptimizer") == "integer" or s.is_integer
        )

        if minimum is None:
            minimum = s.minimum
            if s.minimum is not None:
                exclusive_minimum = exclusive_minimum or s.exclusive_minimum
        elif s.minimum is not None and minimum == s.minimum:
            exclusive_minimum = exclusive_minimum or s.exclusive_minimum

        if maximum is None:
            maximum = s.maximum
            if s.maximum is not None:
                exclusive_maximum = exclusive_maximum or s.exclusive_maximum
        elif s.maximum is not None and minimum == s.minimum:
            exclusive_maximum = exclusive_maximum or s.exclusive_maximum

        return SchemaRange(
            minimum=minimum,
            maximum=maximum,
            exclusive_minimum=exclusive_minimum,
            exclusive_maximum=exclusive_maximum,
            is_integer=is_integer,
        )

    @classmethod
    def to_schema_with_optimizer(
        cls, actual_range: "SchemaRange", optimizer_range: "SchemaRange"
    ) -> JsonSchema:
        number_schema: JsonSchema = {}
        if actual_range.is_integer:
            number_schema["type"] = "integer"
        else:
            number_schema["type"] = "number"
            if optimizer_range.is_integer:
                number_schema["laleType"] = "integer"

        if actual_range.minimum is not None:
            number_schema["minimum"] = actual_range.minimum
            if actual_range.exclusive_minimum:
                number_schema["exclusiveMinimum"] = True

        if optimizer_range.minimum is not None:
            if (
                actual_range.minimum is None
                or actual_range.minimum < optimizer_range.minimum
                or (
                    actual_range.minimum == optimizer_range.minimum
                    and optimizer_range.exclusive_minimum
                    and not actual_range.exclusive_minimum
                )
            ):
                number_schema["minimumForOptimizer"] = optimizer_range.minimum
                if optimizer_range.exclusive_minimum:
                    number_schema["exclusiveMinimumForOptimizer"] = True

        if actual_range.maximum is not None:
            number_schema["maximum"] = actual_range.maximum
            if actual_range.exclusive_maximum:
                number_schema["exclusiveMaximum"] = True

        if optimizer_range.maximum is not None:
            if (
                actual_range.maximum is None
                or actual_range.maximum > optimizer_range.maximum
                or (
                    actual_range.maximum == optimizer_range.maximum
                    and optimizer_range.exclusive_maximum
                    and not actual_range.exclusive_maximum
                )
            ):
                number_schema["maximumForOptimizer"] = optimizer_range.maximum
                if optimizer_range.exclusive_maximum:
                    number_schema["exclusiveMaximumForOptimizer"] = True

        return number_schema

    def __iand__(self, other: "SchemaRange"):
        self.is_integer = self.is_integer or other.is_integer
        if other.minimum is not None:
            if self.minimum is None:
                self.minimum = other.minimum
                self.exclusive_minimum = other.exclusive_minimum
            elif self.minimum == other.minimum:
                self.exclusive_minimum = (
                    self.exclusive_minimum or other.exclusive_minimum
                )
            elif self.minimum < other.minimum:
                self.minimum = other.minimum
                self.exclusive_minimum = other.exclusive_minimum

        if other.maximum is not None:
            if self.maximum is None:
                self.maximum = other.maximum
                self.exclusive_maximum = other.exclusive_maximum
            elif self.maximum == other.maximum:
                self.exclusive_maximum = (
                    self.exclusive_maximum or other.exclusive_maximum
                )
            elif self.maximum > other.maximum:
                self.maximum = other.maximum
                self.exclusive_maximum = other.exclusive_maximum

        return self

    def diff(self, other: "SchemaRange") -> Optional[bool]:
        """Returns None if the resulting region is impossible.
        Returns True if the other constraint was completely subtracted from
        self.  If it could not be, then it returns False (and the caller should probably
        keep the other constraint as a negated constraint)
        """
        # for now, just handle simple exclusions
        if not other.is_integer or other.is_integer == self.is_integer:
            # the exclusion is less than the actual range
            if (
                self.minimum is not None
                and other.maximum is not None
                and (
                    other.maximum < self.minimum
                    or (
                        other.maximum == self.minimum
                        and (self.exclusive_minimum or other.exclusive_maximum)
                    )
                )
            ):
                return True
            # the exclusion is greater than the actual range
            if (
                self.maximum is not None
                and other.minimum is not None
                and (
                    other.minimum > self.maximum
                    or (
                        other.minimum == self.maximum
                        and (self.exclusive_maximum or other.exclusive_minimum)
                    )
                )
            ):
                return True

            if other.minimum is None:
                if self.minimum is None:
                    # the exclusion and the range have no minimum
                    if other.maximum is None:
                        # nothing is possible
                        return None
                    else:
                        self.minimum = other.maximum
                        self.exclusive_minimum = not other.exclusive_maximum
                        return True
                # else might create a hole, so avoid this case
            else:
                # ASSERT: other.minimum is not None
                if (
                    self.minimum is None
                    or self.minimum < other.minimum
                    or (
                        self.minimum == other.minimum
                        and (not self.exclusive_minimum or other.exclusive_minimum)
                    )
                ):
                    if (
                        other.maximum is None
                        or self.maximum is not None
                        and (
                            other.maximum > self.maximum
                            or (
                                (
                                    other.maximum == self.maximum
                                    and (
                                        not other.exclusive_maximum
                                        or self.exclusive_maximum
                                    )
                                )
                            )
                        )
                    ):
                        self.maximum = other.minimum
                        self.exclusive_maximum = not other.exclusive_minimum
                        return True
                    # else might create a hole, so avoid this case
                else:
                    # self.minimum >= other.minimum
                    if (
                        other.maximum is None
                        or self.maximum < other.maximum
                        or (
                            self.maximum == other.maximum
                            and (not other.exclusive_maximum or self.exclusive_maximum)
                        )
                    ):
                        # nothing is possible
                        return None
                    else:
                        self.minimum = other.maximum
                        self.exclusive_minimum = not other.exclusive_maximum
                        return True

            if other.maximum is None:
                if self.maximum is None:
                    # the exclusion and the range have no maximum
                    if other.minimum is None:
                        # nothing is possible
                        return None
                    else:
                        self.maximum = other.minimum
                        self.exclusive_maximum = not other.exclusive_minimum
                        return True
                # else might create a hole, so avoid this case
            else:
                # ASSERT: other.maximum is not None
                if (
                    self.maximum is None
                    or self.maximum > other.maximum
                    or (
                        self.maximum == other.maximum
                        and (not self.exclusive_maximum or other.exclusive_maximum)
                    )
                ):
                    if (
                        other.minimum is None
                        or self.minimum is not None
                        and (
                            other.minimum < self.minimum
                            or (
                                (
                                    other.minimum == self.minimum
                                    and (
                                        not other.exclusive_minimum
                                        or self.exclusive_minimum
                                    )
                                )
                            )
                        )
                    ):
                        self.minimum = other.maximum
                        self.exclusive_minimum = not other.exclusive_maximum
                        return True
                    # else might create a hole, so avoid this case
                else:
                    # self.maximum >= other.maximum
                    if (
                        other.minimum is None
                        or self.minimum > other.minimum
                        or (
                            self.minimum == other.minimum
                            and (not other.exclusive_minimum or self.exclusive_minimum)
                        )
                    ):
                        # nothing is possible
                        return None
                    else:
                        self.maximum = other.minimum
                        self.exclusive_maximum = not other.exclusive_minimum
                        return True
        return False

    def remove_point(self, other: Union[int, float]) -> Optional[bool]:
        """Returns None if the resulting region is impossible.
        Returns True if the other constraint was completely subtracted from
        self.  If it could not be, then it returns False (and the caller should probably
        keep the other constraint as a negated constraint)
        """
        return self.diff(SchemaRange.point(other))

    @classmethod
    def is_empty2(cls, lower: "SchemaRange", upper: "SchemaRange") -> bool:
        """Determines if the range given by taking lower bounds from lower and upper bound from upper is empty (contains nothing)
        is_integer is assumed to be their disjunction
        """
        is_integer = lower.is_integer or upper.is_integer

        if lower.minimum is not None and upper.maximum is not None:
            if lower.minimum > upper.maximum:
                return True
            if lower.minimum == upper.maximum and (
                lower.exclusive_minimum or upper.exclusive_maximum
            ):
                return True
            if (
                is_integer
                and lower.exclusive_minimum
                and upper.exclusive_maximum
                and lower.minimum + 1 == upper.maximum
            ):
                return True
        return False

    def is_empty(self) -> bool:
        """Determines if the range is empty (contains nothing)"""
        return SchemaRange.is_empty2(self, self)



from numpy import inf, nan
from sklearn.preprocessing import Normalizer as Op

from lale.docstrings import set_docstrings
from lale.operators import make_operator


class _NormalizerImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)


_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "inherited docstring for Normalizer    Normalize samples individually to unit norm.",
    "allOf": [
        {
            "type": "object",
            "required": ["norm", "copy"],
            "relevantToOptimizer": ["norm", "copy"],
            "additionalProperties": False,
            "properties": {
                "norm": {
                    "XXX TODO XXX": "'l1', 'l2', or 'max', optional ('l2' by default)",
                    "description": "The norm to use to normalize each non zero sample.",
                    "enum": ["l2"],
                    "default": "l2",
                },
                "copy": {
                    "type": "boolean",
                    "default": True,
                    "description": "set to False to perform inplace row normalization and avoid a copy (if the input is already a numpy array or a scipy.sparse CSR matrix).",
                },
            },
        }
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Do nothing and return the estimator unchanged",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
            "XXX TODO XXX": "array-like",
        }
    },
}
_input_transform_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Scale each non zero row of X to unit norm",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "The data to normalize, row by row",
        },
        "y": {"laleType": "Any", "XXX TODO XXX": "(ignored)", "description": ""},
        "copy": {
            "anyOf": [{"type": "boolean"}, {"enum": [None]}],
            "default": None,
            "description": "Copy the input X or not.",
        },
    },
}
_output_transform_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Scale each non zero row of X to unit norm",
    "laleType": "Any",
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://scikit-learn.org/0.20/modules/generated/sklearn.preprocessing.Normalizer#sklearn-preprocessing-normalizer",
    "import_from": "sklearn.preprocessing",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}
Normalizer = make_operator(_NormalizerImpl, _combined_schemas)

set_docstrings(Normalizer)



from numpy import inf, nan
from sklearn.linear_model import RidgeClassifierCV as Op

from lale.docstrings import set_docstrings
from lale.operators import make_operator


class _RidgeClassifierCVImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)

    def decision_function(self, X):
        return self._wrapped_model.decision_function(X)


_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "inherited docstring for RidgeClassifierCV    Ridge classifier with built-in cross-validation.",
    "allOf": [
        {
            "type": "object",
            "required": [
                "alphas",
                "fit_intercept",
                "normalize",
                "scoring",
                "cv",
                "class_weight",
                "store_cv_values",
            ],
            "relevantToOptimizer": [
                "fit_intercept",
                "normalize",
                "scoring",
                "cv",
                "store_cv_values",
                "class_weight",
            ],
            "additionalProperties": False,
            "properties": {
                "alphas": {
                    "type": "array",
                    "items": {"type": "number"},
                    "default": [0.1, 1.0, 10.0],
                    "description": "Array of alpha values to try",
                },
                "fit_intercept": {
                    "type": "boolean",
                    "default": True,
                    "description": "Whether to calculate the intercept for this model",
                },
                "normalize": {
                    "type": "boolean",
                    "default": False,
                    "description": "This parameter is ignored when ``fit_intercept`` is set to False",
                },
                "scoring": {
                    "anyOf": [
                        {"laleType": "callable", "forOptimizer": False},
                        {"enum": ["accuracy", None]},
                    ],
                    "default": None,
                    "description": "A string (see model evaluation documentation) or a scorer callable object / function with signature ``scorer(estimator, X, y)``.",
                },
                "cv": {
                    "description": """Cross-validation as integer or as object that has a split function.
                        The fit method performs cross validation on the input dataset for per
                        trial, and uses the mean cross validation performance for optimization.
                        This behavior is also impacted by handle_cv_failure flag.
                        If integer: number of folds in sklearn.model_selection.StratifiedKFold.
                        If object with split function: generator yielding (train, test) splits
                        as arrays of indices. Can use any of the iterators from
                        https://scikit-learn.org/stable/modules/cross_validation.html#cross-validation-iterators.""",
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimum": 1,
                            "default": 5,
                            "minimumForOptimizer": 3,
                            "maximumForOptimizer": 4,
                            "distribution": "uniform",
                        },
                        {"laleType": "Any", "forOptimizer": False},
                        {"enum": [None]},
                    ],
                    "default": None,
                },
                "class_weight": {
                    "XXX TODO XXX": "dict or 'balanced', optional",
                    "description": "Weights associated with classes in the form ``{class_label: weight}``",
                    "anyOf": [{"enum": ["balanced"]}, {"enum": [None]}],
                    "default": None,
                },
                "store_cv_values": {
                    "type": "boolean",
                    "default": False,
                    "description": "Flag indicating if the cross-validation values corresponding to each alpha should be stored in the ``cv_values_`` attribute (see below)",
                },
            },
        },
        {
            "XXX TODO XXX": "Parameter: store_cv_values > only compatible with cv=none (i"
        },
        {
            "description": "cv!=None and store_cv_values=True are incompatible",
            "anyOf": [
                {"type": "object", "properties": {"cv": {"enum": [None]}}},
                {
                    "type": "object",
                    "properties": {"store_cv_values": {"enum": [False]}},
                },
            ],
        },
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Fit the ridge classifier.",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "Training vectors, where n_samples is the number of samples and n_features is the number of features.",
        },
        "y": {
            "type": "array",
            "items": {"type": "number"},
            "description": "Target values",
        },
        "sample_weight": {
            "anyOf": [
                {"type": "number"},
                {"type": "array", "items": {"type": "number"}},
            ],
            "description": "Sample weight.",
        },
    },
}
_input_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predict class labels for samples in X.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
                    "XXX TODO XXX": "array_like or sparse matrix, shape (n_samples, n_features)",
                },
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
            "description": "Samples.",
        }
    },
}
_output_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predicted class label per sample.",
    "type": "array",
    "items": {"type": "number"},
}
_input_decision_function_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predict confidence scores for samples.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
                    "XXX TODO XXX": "array_like or sparse matrix, shape (n_samples, n_features)",
                },
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
            "description": "Samples.",
        }
    },
}
_output_decision_function_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Confidence scores per (sample, class) combination",
    "laleType": "Any",
    "XXX TODO XXX": "array, shape=(n_samples,) if n_classes == 2 else (n_samples, n_classes)",
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://scikit-learn.org/0.20/modules/generated/sklearn.linear_model.RidgeClassifierCV#sklearn-linear_model-ridgeclassifiercv",
    "import_from": "sklearn.linear_model",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "classifier"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
        "input_decision_function": _input_decision_function_schema,
        "output_decision_function": _output_decision_function_schema,
    },
}
RidgeClassifierCV = make_operator(_RidgeClassifierCVImpl, _combined_schemas)

set_docstrings(RidgeClassifierCV)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import sklearn.datasets
from sklearn.model_selection import train_test_split
from sklearn.utils import shuffle

import lale.datasets.data_schemas


def _bunch_to_df(bunch, schema_X, schema_y, test_size=0.2, random_state=42):
    train_X_arr, test_X_arr, train_y_arr, test_y_arr = train_test_split(
        bunch.data, bunch.target, test_size=test_size, random_state=random_state
    )
    feature_schemas = schema_X["items"]["items"]
    if isinstance(feature_schemas, list):
        feature_names = [f["description"] for f in feature_schemas]
    else:
        feature_names = [f"x{i}" for i in range(schema_X["items"]["maxItems"])]
    train_X_df = pd.DataFrame(train_X_arr, columns=feature_names)
    test_X_df = pd.DataFrame(test_X_arr, columns=feature_names)
    train_y_df = pd.Series(train_y_arr, name="target")
    test_y_df = pd.Series(test_y_arr, name="target")
    train_nrows, test_nrows = train_X_df.shape[0], test_X_df.shape[0]
    train_X = lale.datasets.data_schemas.add_schema(
        train_X_df, {**schema_X, "minItems": train_nrows, "maxItems": train_nrows}
    )
    test_X = lale.datasets.data_schemas.add_schema(
        test_X_df, {**schema_X, "minItems": test_nrows, "maxItems": test_nrows}
    )
    train_y = lale.datasets.data_schemas.add_schema(
        train_y_df, {**schema_y, "minItems": train_nrows, "maxItems": train_nrows}
    )
    test_y = lale.datasets.data_schemas.add_schema(
        test_y_df, {**schema_y, "minItems": test_nrows, "maxItems": test_nrows}
    )
    return (train_X, train_y), (test_X, test_y)


def load_iris_df(test_size=0.2):
    iris = sklearn.datasets.load_iris()
    X = iris.data
    y = iris.target
    target_name = "target"
    X, y = shuffle(iris.data, iris.target, random_state=42)
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=test_size, random_state=42
    )

    X_train_df = pd.DataFrame(X_train, columns=iris.feature_names)
    y_train_df = pd.Series(y_train, name=target_name)

    X_test_df = pd.DataFrame(X_test, columns=iris.feature_names)
    y_test_df = pd.Series(y_test, name=target_name)

    return (X_train_df, y_train_df), (X_test_df, y_test_df)


def digits_df(test_size=0.2, random_state=42):
    digits = sklearn.datasets.load_digits()
    ncols = digits.data.shape[1]
    schema_X = {
        "description": "Features of digits dataset (classification).",
        "documentation_url": "https://scikit-learn.org/0.20/datasets/index.html#optical-recognition-of-handwritten-digits-dataset",
        "type": "array",
        "items": {
            "type": "array",
            "minItems": ncols,
            "maxItems": ncols,
            "items": {"type": "number", "minimum": 0, "maximum": 16},
        },
    }
    schema_y = {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "type": "array",
        "items": {"type": "integer", "minimum": 0, "maximum": 9},
    }
    (train_X, train_y), (test_X, test_y) = _bunch_to_df(
        digits, schema_X, schema_y, test_size, random_state
    )
    return (train_X, train_y), (test_X, test_y)


def covtype_df(test_size=0.2, random_state=42):
    covtype = sklearn.datasets.fetch_covtype()
    schema_X = {
        "description": "Features of forest covertypes dataset (classification).",
        "documentation_url": "https://scikit-learn.org/0.20/datasets/index.html#forest-covertypes",
        "type": "array",
        "items": {
            "type": "array",
            "minItems": 54,
            "maxItems": 54,
            "items": [
                {"description": "Elevation", "type": "integer"},
                {"description": "Aspect", "type": "integer"},
                {"description": "Slope", "type": "integer"},
                {"description": "Horizontal_Distance_To_Hydrology", "type": "integer"},
                {"description": "Vertical_Distance_To_Hydrology", "type": "integer"},
                {"description": "Horizontal_Distance_To_Roadways", "type": "integer"},
                {"description": "Hillshade_9am", "type": "integer"},
                {"description": "Hillshade_Noon", "type": "integer"},
                {"description": "Hillshade_3pm", "type": "integer"},
                {
                    "description": "Horizontal_Distance_To_Fire_Points",
                    "type": "integer",
                },
                {"description": "Wilderness_Area1", "enum": [0, 1]},
                {"description": "Wilderness_Area2", "enum": [0, 1]},
                {"description": "Wilderness_Area3", "enum": [0, 1]},
                {"description": "Wilderness_Area4", "enum": [0, 1]},
                {"description": "Soil_Type1", "enum": [0, 1]},
                {"description": "Soil_Type2", "enum": [0, 1]},
                {"description": "Soil_Type3", "enum": [0, 1]},
                {"description": "Soil_Type4", "enum": [0, 1]},
                {"description": "Soil_Type5", "enum": [0, 1]},
                {"description": "Soil_Type6", "enum": [0, 1]},
                {"description": "Soil_Type7", "enum": [0, 1]},
                {"description": "Soil_Type8", "enum": [0, 1]},
                {"description": "Soil_Type9", "enum": [0, 1]},
                {"description": "Soil_Type10", "enum": [0, 1]},
                {"description": "Soil_Type11", "enum": [0, 1]},
                {"description": "Soil_Type12", "enum": [0, 1]},
                {"description": "Soil_Type13", "enum": [0, 1]},
                {"description": "Soil_Type14", "enum": [0, 1]},
                {"description": "Soil_Type15", "enum": [0, 1]},
                {"description": "Soil_Type16", "enum": [0, 1]},
                {"description": "Soil_Type17", "enum": [0, 1]},
                {"description": "Soil_Type18", "enum": [0, 1]},
                {"description": "Soil_Type19", "enum": [0, 1]},
                {"description": "Soil_Type20", "enum": [0, 1]},
                {"description": "Soil_Type21", "enum": [0, 1]},
                {"description": "Soil_Type22", "enum": [0, 1]},
                {"description": "Soil_Type23", "enum": [0, 1]},
                {"description": "Soil_Type24", "enum": [0, 1]},
                {"description": "Soil_Type25", "enum": [0, 1]},
                {"description": "Soil_Type26", "enum": [0, 1]},
                {"description": "Soil_Type27", "enum": [0, 1]},
                {"description": "Soil_Type28", "enum": [0, 1]},
                {"description": "Soil_Type29", "enum": [0, 1]},
                {"description": "Soil_Type30", "enum": [0, 1]},
                {"description": "Soil_Type31", "enum": [0, 1]},
                {"description": "Soil_Type32", "enum": [0, 1]},
                {"description": "Soil_Type33", "enum": [0, 1]},
                {"description": "Soil_Type34", "enum": [0, 1]},
                {"description": "Soil_Type35", "enum": [0, 1]},
                {"description": "Soil_Type36", "enum": [0, 1]},
                {"description": "Soil_Type37", "enum": [0, 1]},
                {"description": "Soil_Type38", "enum": [0, 1]},
                {"description": "Soil_Type39", "enum": [0, 1]},
                {"description": "Soil_Type40", "enum": [0, 1]},
            ],
        },
    }
    schema_y = {
        "description": "Target of forest covertypes dataset (classification).",
        "documentation_url": "https://scikit-learn.org/0.20/datasets/index.html#forest-covertypes",
        "type": "array",
        "items": {
            "description": "The cover type, i.e., the dominant species of trees.",
            "enum": [0, 1, 2, 3, 4, 5, 6],
        },
    }
    (train_X, train_y), (test_X, test_y) = _bunch_to_df(
        covtype, schema_X, schema_y, test_size, random_state
    )
    return (train_X, train_y), (test_X, test_y)


def california_housing_df(test_size=0.2, random_state=42):
    housing = sklearn.datasets.fetch_california_housing()
    schema_X = {
        "description": "Features of California housing dataset (regression).",
        "documentation_url": "https://scikit-learn.org/0.20/datasets/index.html#california-housing-dataset",
        "type": "array",
        "items": {
            "type": "array",
            "minItems": 8,
            "maxItems": 8,
            "items": [
                {"description": "MedInc", "type": "number", "minimum": 0.0},
                {"description": "HouseAge", "type": "number", "minimum": 0.0},
                {"description": "AveRooms", "type": "number", "minimum": 0.0},
                {"description": "AveBedrms", "type": "number", "minimum": 0.0},
                {"description": "Population", "type": "number", "minimum": 0.0},
                {"description": "AveOccup", "type": "number", "minimum": 0.0},
                {"description": "Latitude", "type": "number", "minimum": 0.0},
                {"description": "Longitude", "type": "number"},
            ],
        },
    }
    schema_y = {
        "description": "Target of California housing dataset (regression).",
        "documentation_url": "https://scikit-learn.org/0.20/datasets/index.html#california-housing-dataset",
        "type": "array",
        "items": {
            "description": "Median house value for California districts.",
            "type": "number",
            "minimum": 0.0,
        },
    }
    (train_X, train_y), (test_X, test_y) = _bunch_to_df(
        housing, schema_X, schema_y, test_size, random_state
    )
    return (train_X, train_y), (test_X, test_y)


def boston_housing_df(test_size=0.2, random_state=42):
    housing = sklearn.datasets.load_boston()
    schema_X = {
        "description": "Features of Boston house prices dataset (regression).",
        "documentation_url": "https://scikit-learn.org/0.20/datasets/index.html#boston-house-prices-dataset",
        "type": "array",
        "items": {
            "type": "array",
            "minItems": 13,
            "maxItems": 13,
            "items": [
                {"description": "CRIM", "type": "number", "minimum": 0.0},
                {"description": "ZN", "type": "number", "minimum": 0.0},
                {"description": "INDUS", "type": "number", "minimum": 0.0},
                {"description": "CHAS", "enum": [0, 1]},
                {"description": "NOX", "type": "number", "minimum": 0.0},
                {"description": "RM", "type": "number", "minimum": 1.0},
                {"description": "AGE", "type": "number", "minimum": 0.0},
                {"description": "DIS", "type": "number", "minimum": 0.0},
                {"description": "RAD", "type": "number", "minimum": 1},
                {"description": "TAX", "type": "number", "minimum": 0.0},
                {"description": "PRATIO", "type": "number", "minimum": 0.0},
                {"description": "B", "type": "number", "minimum": 0.0},
                {"description": "LSTAT", "type": "number", "minimum": 0.0},
            ],
        },
    }

    schema_y = {
        "description": "Target of Boston house prices dataset (regression).",
        "documentation_url": "https://scikit-learn.org/0.20/datasets/index.html#boston-house-prices-dataset",
        "type": "array",
        "items": {
            "description": "Median value of owner-occupied homes in $1000's (MEDV)",
            "type": "number",
            "minimum": 0.0,
        },
    }

    (train_X, train_y), (test_X, test_y) = _bunch_to_df(
        housing, schema_X, schema_y, test_size, random_state
    )
    return (train_X, train_y), (test_X, test_y)



# Copyright 2020 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import autoai_libs.transformers.exportable
import numpy as np
import pandas as pd

import lale.datasets.data_schemas
import lale.docstrings
import lale.operators


class _boolean2floatImpl:
    def __init__(self, activate_flag=True):
        self._hyperparams = {"activate_flag": activate_flag}
        self._wrapped_model = autoai_libs.transformers.exportable.boolean2float(
            **self._hyperparams
        )

    def fit(self, X, y=None):
        self._wrapped_model.fit(X, y)
        return self

    def transform(self, X):
        raw = self._wrapped_model.transform(X)
        if isinstance(raw, np.ndarray) or isinstance(raw, pd.DataFrame):
            s_X = lale.datasets.data_schemas.to_schema(X)
            s_result = self.transform_schema(s_X)
            result = lale.datasets.data_schemas.add_schema(raw, s_result, recalc=True)
        else:
            result = raw
        return result

    def transform_schema(self, s_X):
        """Used internally by Lale for type-checking downstream operators."""
        if self._hyperparams["activate_flag"]:
            result = {
                "type": "array",
                "items": {"type": "array", "items": {"type": "number"}},
            }
        else:
            result = s_X
        return result


_hyperparams_schema = {
    "allOf": [
        {
            "description": "This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters.",
            "type": "object",
            "additionalProperties": False,
            "required": ["activate_flag"],
            "relevantToOptimizer": [],
            "properties": {
                "activate_flag": {
                    "description": "If False, transform(X) outputs the input numpy array X unmodified.",
                    "type": "boolean",
                    "default": True,
                }
            },
        }
    ]
}

_input_fit_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {  # Handles 1-D arrays as well
            "anyOf": [
                {"type": "array", "items": {"laleType": "Any"}},
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"laleType": "Any"}},
                },
            ]
        },
        "y": {"laleType": "Any"},
    },
}

_input_transform_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {  # Handles 1-D arrays as well
            "anyOf": [
                {"type": "array", "items": {"laleType": "Any"}},
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"laleType": "Any"}},
                },
            ]
        }
    },
}

_output_transform_schema = {
    "description": "Features; the outer array is over samples.",
    "type": "array",
    "items": {"type": "array", "items": {"laleType": "Any"}},
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """Operator from `autoai_libs`_. Converts strings that represent booleans to floats and replaces missing values with np.nan.

.. _`autoai_libs`: https://pypi.org/project/autoai-libs""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.autoai_libs.boolean2float.html",
    "import_from": "autoai_libs.transformers.exportable",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}


boolean2float = lale.operators.make_operator(_boolean2floatImpl, _combined_schemas)

lale.docstrings.set_docstrings(boolean2float)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from torch.utils.data import Dataset
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        """Your Python environment does not have torch installed. You can install it with
                                pip install torch
                                or with
                                    pip install 'lale[full]'"""
    )
try:
    import h5py
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        """Your Python environment does not have h5py installed. You can install it with
                                pip install h5py
                                or with
                                    pip install 'lale[full]'"""
    )


class HDF5TorchDataset(Dataset):
    """Pytorch Dataset subclass that takes a hdf5 file pointer."""

    def __init__(self, file_path):
        """.

        Parameters
        ----------
        file : file is an object of class h5py.File
        """
        self.file_path = file_path
        h5_file = h5py.File(file_path)
        self.length = h5_file["X"].shape[0]

    def __len__(self):
        return self.length

    def __getitem__(self, idx):
        with h5py.File(self.file_path) as h5_file:
            X = h5_file["X"]
            try:
                y = h5_file["y"]
            except KeyError:
                y = None
            if y is None:
                element = X[idx]
            else:
                element = X[idx], y[idx]
        return element

    def get_data(self):
        with h5py.File(self.file_path) as h5_file:
            X = h5_file["X"][:]
            try:
                y = h5_file["y"][:]
            except KeyError:
                y = None
            if y is None:
                return X
            else:
                return X, y



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import logging
from typing import Any, Dict

from hyperopt import STATUS_OK

import lale.docstrings
import lale.operators
from lale.lib.lale import Hyperopt

logger = logging.getLogger(__name__)


class _TopKVotingClassifierImpl:
    args_to_optimizer: Dict[str, Any]

    def __init__(
        self,
        estimator=None,
        optimizer=None,
        args_to_optimizer: Dict[str, Any] = None,
        k=10,
    ):
        self.estimator = estimator
        if self.estimator is None:
            raise ValueError("Estimator is a required argument.")
        self.optimizer = optimizer
        if self.optimizer is None:
            self.optimizer = Hyperopt
        if args_to_optimizer is None:
            self.args_to_optimizer = {}
        else:
            self.args_to_optimizer = args_to_optimizer
        self.k = k

    def fit(self, X_train, y_train, **fit_params):
        from lale.lib.sklearn import VotingClassifier

        optimizer_instance = self.optimizer(
            estimator=self.estimator, **self.args_to_optimizer
        )
        trained_optimizer1 = optimizer_instance.fit(X_train, y_train, **fit_params)
        results = trained_optimizer1.summary()
        results = results[
            results["status"] == STATUS_OK
        ]  # Consider only successful trials
        results = results.sort_values(by=["loss"], axis=0)
        k = min(self.k, results.shape[0])
        top_k_pipelines = results.iloc[0:k]
        pipeline_tuples = []
        for pipeline_name in top_k_pipelines.index:
            pipeline_instance = trained_optimizer1.get_pipeline(pipeline_name)
            pipeline_tuple = (pipeline_name, pipeline_instance)
            pipeline_tuples.append(pipeline_tuple)
        voting = VotingClassifier(estimators=pipeline_tuples)
        args_to_optimizer = copy.copy(self.args_to_optimizer)
        try:
            del args_to_optimizer["max_evals"]
        except KeyError:
            pass
        args_to_optimizer[
            "max_evals"
        ] = 1  # Currently, voting classifier has no useful hyperparameters to tune.
        optimizer_instance2 = self.optimizer(estimator=voting, **args_to_optimizer)
        trained_optimizer2 = optimizer_instance2.fit(X_train, y_train, **fit_params)
        self._best_estimator = trained_optimizer2.get_pipeline()
        return self

    def predict(self, X_eval, **predict_params):
        import warnings

        warnings.filterwarnings("ignore")
        if self._best_estimator is None:
            raise ValueError(
                "Can not predict as the best estimator is None. Either an attempt to call `predict` "
                "before calling `fit` or all the trials during `fit` failed."
            )
        trained = self._best_estimator
        try:
            predictions = trained.predict(X_eval, **predict_params)
        except ValueError as e:
            logger.warning(
                "ValueError in predicting using Hyperopt:{}, the error is:{}".format(
                    trained, e
                )
            )
            predictions = None
        return predictions

    def get_pipeline(self, pipeline_name=None, astype="lale"):
        """Retrieve one of the trials.

        Parameters
        ----------
        pipeline_name : None

        astype : 'lale' or 'sklearn', default 'lale'
            Type of resulting pipeline.

        Returns
        -------
        result : Trained operator if best, trainable operator otherwise."""
        assert pipeline_name is None
        result = self._best_estimator
        if result is None or astype == "lale":
            return result
        assert astype == "sklearn", astype
        assert isinstance(result, lale.operators.BasePipeline)
        return result.export_to_sklearn_pipeline()


_hyperparams_schema = {
    "allOf": [
        {
            "type": "object",
            "required": ["estimator"],
            "relevantToOptimizer": [],
            "additionalProperties": False,
            "properties": {
                "estimator": {
                    "description": "Planned Lale individual operator or pipeline.",
                    "anyOf": [
                        {"laleType": "operator", "not": {"enum": [None]}},
                        {"enum": [None]},
                    ],
                    "default": None,
                },
                "optimizer": {
                    "description": """Optimizer class to be used during the two stages of optimization.
                Default of None uses Hyperopt internally. Currently, only Hyperopt is supported as an optimizer.""",
                    "anyOf": [
                        {"laleType": "operator", "not": {"enum": [None]}},
                        {"enum": [None]},
                    ],
                    "default": None,
                },
                "args_to_optimizer": {
                    "description": """Dictionary of keyword arguments required to be used for the given optimizer
                                as applicable for the given task. For example, max_evals, cv, scoring etc. for Hyperopt.
                                If None, default values for the optimizer would be used.""",
                    "anyOf": [
                        {"type": "object"},  # Python dictionary
                        {"enum": [None]},
                    ],
                    "default": None,
                },
                "k": {
                    "description": """Number of top pipelines to be used for the voting ensemble. If the number of
                            successful trials of the optimizer are less than k, the ensemble will use
                            only successful trials.""",
                    "type": "integer",
                    "minimum": 1,
                    "default": 10,
                },
            },
        }
    ]
}

_input_fit_schema = {
    "type": "object",
    "required": ["X", "y"],
    "properties": {"X": {}, "y": {}},
}
_input_predict_schema = {"type": "object", "required": ["X"], "properties": {"X": {}}}

_output_predict_schema: Dict[str, Any] = {}

_combined_schemas = {
    "description": """This operator creates a voting ensemble from top k performing pipelines from the given planned pipeline.""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.lale.topk_voting_classifier.html",
    "import_from": "lale.lib.lale",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
    },
}


TopKVotingClassifier = lale.operators.make_operator(
    _TopKVotingClassifierImpl, _combined_schemas
)

lale.docstrings.set_docstrings(TopKVotingClassifier)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, TypeVar, Union


class Undefined:
    pass


undefined = Undefined()

T = TypeVar("T")
Option = Union[Undefined, T]


class Schema:

    schema: Dict[str, Any]

    def __init__(
        self,
        desc: Option[str] = undefined,
        default: Option[Any] = undefined,
        forOptimizer: bool = True,
    ):
        self.schema: Dict[str, Any] = {}
        if not isinstance(default, Undefined):
            self.schema["default"] = default
        if not isinstance(desc, Undefined):
            self.schema["description"] = desc
        if not forOptimizer:
            self.schema["forOptimizer"] = forOptimizer

    def set(self, prop: str, value: Option[Any]):
        if not isinstance(value, Undefined):
            self.schema[prop] = value


# Base Type


class Bool(Schema):
    def __init__(
        self,
        desc: Option[str] = undefined,
        default: Option[bool] = undefined,
        forOptimizer: bool = True,
    ):
        super().__init__(desc, default, forOptimizer)
        self.set("type", "boolean")


class Enum(Schema):
    def __init__(
        self,
        values: List[Any] = [],
        desc: Option[str] = undefined,
        default: Option[Any] = undefined,
        forOptimizer: bool = True,
    ):
        super().__init__(desc, default, forOptimizer)
        self.set("enum", values)


class Float(Schema):
    def __init__(
        self,
        desc: Option[str] = undefined,
        default: Option[float] = undefined,
        forOptimizer: bool = True,
        minimum: Option[float] = undefined,
        exclusiveMinimum: Option[bool] = undefined,
        minimumForOptimizer: Option[float] = undefined,
        exclusiveMinimumForOptimizer: Option[bool] = undefined,
        maximum: Option[float] = undefined,
        exclusiveMaximum: Option[bool] = undefined,
        maximumForOptimizer: Option[float] = undefined,
        exclusiveMaximumForOptimizer: Option[bool] = undefined,
        distribution: Option[str] = undefined,
    ):
        super().__init__(desc, default, forOptimizer)
        self.set("type", "number")
        self.set("minimum", minimum)
        self.set("exclusiveMinimum", exclusiveMinimum)
        self.set("minimumForOptimizer", minimumForOptimizer)
        self.set("exclusiveMinimumForOptimizer", exclusiveMinimumForOptimizer)
        self.set("maximum", maximum)
        self.set("exclusiveMaximum", exclusiveMaximum)
        self.set("maximumForOptimizer", maximumForOptimizer)
        self.set("exclusiveMaximumForOptimizer", exclusiveMaximumForOptimizer)
        self.set("distribution", distribution)


class Int(Schema):
    def __init__(
        self,
        desc: Option[str] = undefined,
        default: Option[int] = undefined,
        forOptimizer: bool = True,
        minimum: Option[int] = undefined,
        exclusiveMinimum: Option[bool] = undefined,
        minimumForOptimizer: Option[int] = undefined,
        exclusiveMinimumForOptimizer: Option[bool] = undefined,
        maximum: Option[int] = undefined,
        exclusiveMaximum: Option[bool] = undefined,
        maximumForOptimizer: Option[int] = undefined,
        exclusiveMaximumForOptimizer: Option[bool] = undefined,
        distribution: Option[str] = undefined,
        laleMaximum: Option[str] = undefined,
    ):
        super().__init__(desc, default, forOptimizer)
        self.set("type", "integer")
        self.set("minimum", minimum)
        self.set("exclusiveMinimum", exclusiveMinimum)
        self.set("minimumForOptimizer", minimumForOptimizer)
        self.set("maximum", maximum)
        self.set("exclusiveMaximum", exclusiveMaximum)
        self.set("exclusiveMinimumForOptimizer", exclusiveMinimumForOptimizer)
        self.set("maximumForOptimizer", maximumForOptimizer)
        self.set("exclusiveMaximumForOptimizer", exclusiveMaximumForOptimizer)
        self.set("distribution", distribution)
        self.set("laleMaximum", laleMaximum)


class Null(Schema):
    def __init__(self, desc: Option[str] = undefined, forOptimizer: bool = True):
        super().__init__(desc=desc, forOptimizer=forOptimizer)
        self.set("enum", [None])


class Not(Schema):
    def __init__(self, body: Schema):
        super().__init__()
        self.schema = {"not": body.schema}


class JSON(Schema):
    def __init__(self, body: Dict[str, Any]):
        super().__init__()
        self.schema = body


# Combinator


class AnyOf(Schema):
    def __init__(
        self,
        types: List[Schema] = [],
        desc: Option[str] = undefined,
        default: Option[Any] = undefined,
        forOptimizer: bool = True,
    ):
        super().__init__(desc, default, forOptimizer)
        self.set("anyOf", [t.schema for t in types])


class AllOf(Schema):
    def __init__(
        self,
        types: List[Schema] = [],
        desc: Option[str] = undefined,
        default: Option[Any] = undefined,
    ):
        super().__init__(desc, default)
        self.set("allOf", [t.schema for t in types])


class Array(Schema):
    def __init__(
        self,
        items: Schema,
        desc: Option[str] = undefined,
        default: Option[List[Any]] = undefined,
        forOptimizer: bool = True,
        minItems: Option[int] = undefined,
        minItemsForOptimizer: Option[int] = undefined,
        maxItems: Option[int] = undefined,
        maxItemsForOptimizer: Option[int] = undefined,
        laleType: Option[str] = undefined,
    ):
        super().__init__(desc, default, forOptimizer)
        self.set("type", "array")
        self.set("items", items.schema)
        self.set("minItems", minItems)
        self.set("minItemsForOptimizer", minItemsForOptimizer)
        self.set("maxItems", maxItems)
        self.set("maxItemsForOptimizer", maxItemsForOptimizer)
        self.set("laleType", laleType)


class Object(Schema):
    def __init__(
        self,
        default: Option[Any] = undefined,
        desc: Option[str] = undefined,
        forOptimizer: bool = True,
        required: Option[List[str]] = undefined,
        additionalProperties: Option[bool] = undefined,
        **kwargs: Schema
    ):
        super().__init__(desc, default, forOptimizer)
        self.set("type", "object")
        self.set("required", required)
        self.set("additionalProperties", additionalProperties)
        self.set("properties", {k: p.schema for (k, p) in kwargs.items()})


class String(Schema):
    def __init__(
        self,
        desc: Option[str] = undefined,
        default: Option[str] = undefined,
        forOptimizer: bool = False,
    ):
        super().__init__(desc, default, forOptimizer)
        self.set("type", "string")



from numpy import inf, nan
from sklearn.decomposition import LatentDirichletAllocation as Op

from lale.docstrings import set_docstrings
from lale.operators import make_operator


class _LatentDirichletAllocationImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)


_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "inherited docstring for LatentDirichletAllocation    Latent Dirichlet Allocation with online variational Bayes algorithm",
    "allOf": [
        {
            "type": "object",
            "required": [
                "n_components",
                "doc_topic_prior",
                "topic_word_prior",
                "learning_method",
                "learning_decay",
                "learning_offset",
                "max_iter",
                "batch_size",
                "evaluate_every",
                "total_samples",
                "perp_tol",
                "mean_change_tol",
                "max_doc_update_iter",
                "n_jobs",
                "verbose",
                "random_state",
            ],
            "relevantToOptimizer": [
                "n_components",
                "max_iter",
                "batch_size",
                "evaluate_every",
                "total_samples",
                "max_doc_update_iter",
                "learning_method",
            ],
            "additionalProperties": False,
            "properties": {
                "n_components": {
                    "type": "integer",
                    "minimun": 1,
                    "laleMaximum": "X/items/maxItems",
                    "minimumForOptimizer": 2,
                    "maximumForOptimizer": 256,
                    "distribution": "uniform",
                    "default": 10,
                    "description": "Number of topics.",
                },
                "doc_topic_prior": {
                    "anyOf": [{"type": "number"}, {"enum": [None]}],
                    "default": None,
                    "description": "Prior of document topic distribution `theta`",
                },
                "topic_word_prior": {
                    "anyOf": [{"type": "number"}, {"enum": [None]}],
                    "default": None,
                    "description": "Prior of topic word distribution `beta`",
                },
                "learning_method": {
                    "enum": ["batch", "online"],
                    "default": "batch",
                    "description": "Method used to update `_component`",
                },
                "learning_decay": {
                    "type": "number",
                    "default": 0.7,
                    "description": "It is a parameter that control learning rate in the online learning method",
                },
                "learning_offset": {
                    "type": "number",
                    "default": 10.0,
                    "description": "A (positive) parameter that downweights early iterations in online learning",
                },
                "max_iter": {
                    "type": "integer",
                    "minimumForOptimizer": 10,
                    "maximumForOptimizer": 1000,
                    "distribution": "uniform",
                    "default": 10,
                    "description": "The maximum number of iterations.",
                },
                "batch_size": {
                    "type": "integer",
                    "minimumForOptimizer": 3,
                    "maximumForOptimizer": 128,
                    "distribution": "uniform",
                    "default": 128,
                    "description": "Number of documents to use in each EM iteration",
                },
                "evaluate_every": {
                    "type": "integer",
                    "minimumForOptimizer": (-1),
                    "maximumForOptimizer": 0,
                    "distribution": "uniform",
                    "default": (-1),
                    "description": "How often to evaluate perplexity",
                },
                "total_samples": {
                    "anyOf": [
                        {"type": "integer", "forOptimizer": False},
                        {
                            "type": "number",
                            "minimumForOptimizer": 0.0,
                            "maximumForOptimizer": 1.0,
                            "distribution": "uniform",
                        },
                    ],
                    "default": 1000000.0,
                    "description": "Total number of documents",
                },
                "perp_tol": {
                    "type": "number",
                    "default": 0.1,
                    "description": "Perplexity tolerance in batch learning",
                },
                "mean_change_tol": {
                    "type": "number",
                    "default": 0.001,
                    "description": "Stopping tolerance for updating document topic distribution in E-step.",
                },
                "max_doc_update_iter": {
                    "type": "integer",
                    "minimumForOptimizer": 100,
                    "maximumForOptimizer": 101,
                    "distribution": "uniform",
                    "default": 100,
                    "description": "Max number of iterations for updating document topic distribution in the E-step.",
                },
                "n_jobs": {
                    "anyOf": [{"type": "integer"}, {"enum": [None]}],
                    "default": 1,
                    "description": "The number of jobs to use in the E-step",
                },
                "verbose": {
                    "type": "integer",
                    "default": 0,
                    "description": "Verbosity level.",
                },
                "random_state": {
                    "anyOf": [
                        {"type": "integer"},
                        {"laleType": "numpy.random.RandomState"},
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "If int, random_state is the seed used by the random number generator; If RandomState instance, random_state is the random number generator; If None, the random number generator is the RandomState instance used by `np.random`.",
                },
            },
        },
        {
            "XXX TODO XXX": "Parameter: perp_tol > only used when evaluate_every is greater than 0"
        },
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Learn model for the data X with variational Bayes method.",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
                    "XXX TODO XXX": "array-like or sparse matrix, shape=(n_samples, n_features)",
                },
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
            "description": "Document word matrix.",
        },
        "y": {},
    },
}
_input_transform_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Transform data X according to the fitted model.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
                    "XXX TODO XXX": "array-like or sparse matrix, shape=(n_samples, n_features)",
                },
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
            "description": "Document word matrix.",
        }
    },
}
_output_transform_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Document topic distribution for X.",
    "laleType": "Any",
    "XXX TODO XXX": "shape=(n_samples, n_components)",
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://scikit-learn.org/0.20/modules/generated/sklearn.decomposition.LatentDirichletAllocation#sklearn-decomposition-latentdirichletallocation",
    "import_from": "sklearn.decomposition",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}
LatentDirichletAllocation = make_operator(
    _LatentDirichletAllocationImpl, _combined_schemas
)

set_docstrings(LatentDirichletAllocation)



# Copyright 2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import numpy as np

import lale.docstrings
import lale.operators
from lale.expressions import it
from lale.expressions import max as agg_max
from lale.expressions import min as agg_min
from lale.helpers import _is_pandas_df, _is_spark_df
from lale.lib.rasl import Aggregate, Map
from lale.lib.sklearn import min_max_scaler
from lale.schemas import Enum


def _df_count(X):
    if _is_pandas_df(X):
        return len(X)
    elif _is_spark_df(X):
        return X.count()


class _MinMaxScalerImpl:
    def __init__(self, feature_range=(0, 1), *, copy=True, clip=False):
        if not copy:
            raise ValueError("`copy=False` is not supported by this implementation")
        if clip:
            raise ValueError("`clip=True` is not supported by this implementation")
        self._hyperparams = {"feature_range": feature_range, "copy": copy, "clip": clip}
        self.n_samples_seen_ = 0
        self._transformer = None

    def fit(self, X, y=None):
        self._set_fit_attributes(self._lift(X, self._hyperparams))
        return self

    def partial_fit(self, X, y=None):
        if self.n_samples_seen_ == 0:  # first fit
            return self.fit(X)
        lifted_a = (
            self.data_min_,
            self.data_max_,
            self.n_samples_seen_,
            self.n_features_in_,
            self.feature_names_in_,
        )
        lifted_b = self._lift(X, self._hyperparams)
        self._set_fit_attributes(self._combine(lifted_a, lifted_b))
        return self

    def transform(self, X):
        if self._transformer is None:
            self._transformer = self._build_transformer(X)
        return self._transformer.transform(X)

    def _set_fit_attributes(self, lifted):
        (
            self.data_min_,
            self.data_max_,
            self.n_samples_seen_,
            self.n_features_in_,
            self.feature_names_in_,
        ) = lifted
        self.data_range_ = self.data_max_ - self.data_min_
        range_min, range_max = self._hyperparams["feature_range"]
        self.scale_ = (range_max - range_min) / (self.data_max_ - self.data_min_)
        self.min_ = range_min - self.data_min_ * self.scale_

    def _build_transformer(self, X):
        range_min, range_max = self._hyperparams["feature_range"]
        ops = {}
        for i, c in enumerate(X.columns):
            c_std = (it[c] - self.data_min_[i]) / (  # type: ignore
                self.data_max_[i] - self.data_min_[i]  # type: ignore
            )
            c_scaled = c_std * (range_max - range_min) + range_min
            ops.update({c: c_scaled})
        return Map(columns=ops)

    @staticmethod
    def _lift(X, hyperparams):
        agg = {f"{c}_min": agg_min(it[c]) for c in X.columns}
        agg.update({f"{c}_max": agg_max(it[c]) for c in X.columns})
        aggregate = Aggregate(columns=agg)
        data_min_max = aggregate.transform(X)
        if _is_spark_df(X):
            data_min_max = data_min_max.toPandas()
        n = len(X.columns)
        data_min_ = np.zeros(shape=(n))
        data_max_ = np.zeros(shape=(n))
        for i, c in enumerate(X.columns):
            data_min_[i] = data_min_max[f"{c}_min"]
            data_max_[i] = data_min_max[f"{c}_max"]
        data_min_ = np.array(data_min_)
        data_max_ = np.array(data_max_)
        n_samples_seen_ = _df_count(X)
        n_features_in_ = len(X.columns)
        feature_names_in_ = X.columns
        return data_min_, data_max_, n_samples_seen_, n_features_in_, feature_names_in_

    @staticmethod
    def _combine(lifted_a, lifted_b):
        (
            data_min_a,
            data_max_a,
            n_samples_seen_a,
            n_features_in_a,
            feature_names_in_a,
        ) = lifted_a
        (
            data_min_b,
            data_max_b,
            n_samples_seen_b,
            n_features_in_b,
            feature_names_in_b,
        ) = lifted_b
        data_min_ = np.minimum(data_min_a, data_min_b)
        data_max_ = np.maximum(data_max_a, data_max_b)
        n_samples_seen_ = n_samples_seen_a + n_samples_seen_b
        assert n_features_in_a == n_features_in_b
        n_features_in_ = n_features_in_a
        assert list(feature_names_in_a) == list(feature_names_in_b)
        feature_names_in_ = feature_names_in_a
        return data_min_, data_max_, n_samples_seen_, n_features_in_, feature_names_in_


_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Relational algebra implementation of MinMaxScaler.",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.min_max_scaler.html",
    "type": "object",
    "tags": {
        "pre": ["~categoricals"],
        "op": ["transformer", "interpretable"],
        "post": [],
    },
    "properties": {
        "hyperparams": min_max_scaler._hyperparams_schema,
        "input_fit": min_max_scaler._input_schema_fit,
        "input_transform": min_max_scaler._input_transform_schema,
        "output_transform": min_max_scaler._output_transform_schema,
    },
}

MinMaxScaler = lale.operators.make_operator(_MinMaxScalerImpl, _combined_schemas)

MinMaxScaler = typing.cast(
    lale.operators.PlannedIndividualOp,
    MinMaxScaler.customize_schema(
        copy=Enum(
            values=[True],
            desc="`copy=True` is the only value currently supported by this implementation",
            default=True,
        ),
        clip=Enum(
            values=[False],
            desc="`clip=False` is the only value currently supported by this implementation",
            default=False,
        ),
    ),
)

lale.docstrings.set_docstrings(MinMaxScaler)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import logging
import multiprocessing
import sys
import time
import traceback
import warnings
from typing import Any, Dict, Optional

import hyperopt
import numpy as np
import pandas as pd
from hyperopt.exceptions import AllTrialsFailed
from sklearn.metrics import check_scoring, log_loss
from sklearn.model_selection import train_test_split
from sklearn.model_selection._split import check_cv

import lale.docstrings
import lale.helpers
import lale.operators
import lale.pretty_print
from lale.helpers import (
    create_instance_from_hyperopt_search_space,
    cross_val_score_track_trials,
)
from lale.lib.sklearn import LogisticRegression
from lale.search.op2hp import hyperopt_search_space
from lale.search.PGO import PGO

from ._common_schemas import (
    schema_best_score_single,
    schema_cv,
    schema_estimator,
    schema_max_opt_time,
    schema_scoring_single,
)

SEED = 42
logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)


class _HyperoptImpl:
    def __init__(
        self,
        *,
        estimator=None,
        scoring=None,
        best_score=0.0,
        args_to_scorer=None,
        cv=5,
        handle_cv_failure=False,
        verbose=False,
        show_progressbar=True,
        algo="tpe",
        max_evals=50,
        frac_evals_with_defaults=0,
        max_opt_time=None,
        max_eval_time=None,
        pgo: Optional[PGO] = None,
    ):
        self.max_evals = max_evals
        if estimator is None:
            self.estimator = LogisticRegression()
        else:
            self.estimator = estimator
        if frac_evals_with_defaults > 0:
            self.evals_with_defaults = int(frac_evals_with_defaults * max_evals)
        else:
            self.evals_with_defaults = 0
        self.algo = algo
        self.scoring = scoring
        if self.scoring is None:
            is_clf = self.estimator.is_classifier()
            if is_clf:
                self.scoring = "accuracy"
            else:
                self.scoring = "r2"
        self.best_score = best_score
        self.handle_cv_failure = handle_cv_failure
        self.cv = cv
        self._trials = hyperopt.Trials()
        self._default_trials = hyperopt.Trials()
        self.max_opt_time = max_opt_time
        self.max_eval_time = max_eval_time
        self.pgo = pgo
        self.show_progressbar = show_progressbar
        if args_to_scorer is not None:
            self.args_to_scorer = args_to_scorer
        else:
            self.args_to_scorer = {}
        self.verbose = verbose

    def _summarize_statuses(self):
        status_list = self._trials.statuses()
        status_hist = {}
        for status in status_list:
            status_hist[status] = 1 + status_hist.get(status, 0)
        if hyperopt.STATUS_FAIL in status_hist:
            print(
                f"{status_hist[hyperopt.STATUS_FAIL]} out of {len(status_list)} trials failed, call summary() for details."
            )
            if not self.verbose:
                print("Run with verbose=True to see per-trial exceptions.")

    def fit(self, X_train, y_train, **fit_params):
        opt_start_time = time.time()
        is_clf = self.estimator.is_classifier()
        self.cv = check_cv(self.cv, y=y_train, classifier=is_clf)
        data_schema = lale.helpers.fold_schema(X_train, y_train, self.cv, is_clf)
        self.search_space = hyperopt.hp.choice(
            "meta_model",
            [
                hyperopt_search_space(
                    self.estimator, pgo=self.pgo, data_schema=data_schema
                )
            ],
        )
        # Create a search space with default hyperparameters for all trainable parts of the pipeline.
        # This search space is used for `frac_evals_with_defaults` fraction of the total trials.
        try:
            self.search_space_with_defaults = hyperopt.hp.choice(
                "meta_model",
                [
                    hyperopt_search_space(
                        self.estimator.freeze_trainable(),
                        pgo=self.pgo,
                        data_schema=data_schema,
                    )
                ],
            )
        except Exception:
            logger.warning(
                "Exception caught during generation of default search space, setting frac_evals_with_defaults to zero."
            )
            self.evals_with_defaults = 0

        def hyperopt_train_test(params, X_train, y_train):
            warnings.filterwarnings("ignore")

            trainable = create_instance_from_hyperopt_search_space(
                self.estimator, params
            )
            try:
                cv_score, logloss, execution_time = cross_val_score_track_trials(
                    trainable,
                    X_train,
                    y_train,
                    cv=self.cv,
                    scoring=self.scoring,
                    args_to_scorer=self.args_to_scorer,
                )
                logger.debug(
                    "Successful trial of hyperopt with hyperparameters:{}".format(
                        params
                    )
                )
            except BaseException as e:
                # If there is any error in cross validation, use the score based on a random train-test split as the evaluation criterion
                if self.handle_cv_failure and trainable is not None:
                    (
                        X_train_part,
                        X_validation,
                        y_train_part,
                        y_validation,
                    ) = train_test_split(X_train, y_train, test_size=0.20)
                    start = time.time()
                    trained = trainable.fit(X_train_part, y_train_part, **fit_params)
                    scorer = check_scoring(trainable, scoring=self.scoring)
                    cv_score = scorer(
                        trained, X_validation, y_validation, **self.args_to_scorer
                    )
                    execution_time = time.time() - start
                    y_pred_proba = trained.predict_proba(X_validation)
                    try:
                        logloss = log_loss(y_true=y_validation, y_pred=y_pred_proba)
                    except BaseException:
                        logloss = 0
                        logger.debug("Warning, log loss cannot be computed")
                else:
                    logger.debug(e)
                    if trainable is None:
                        logger.debug(
                            "Error {} with uncreatable pipeline with parameters:{}".format(
                                e, lale.pretty_print.hyperparams_to_string(params)
                            )
                        )
                    else:
                        logger.debug(
                            "Error {} with pipeline:{}".format(e, trainable.to_json())
                        )
                    raise e
            return cv_score, logloss, execution_time

        def merge_trials(trials1, trials2):
            max_tid = max([trial["tid"] for trial in trials1.trials])

            for trial in trials2:
                tid = trial["tid"] + max_tid + 1
                hyperopt_trial = hyperopt.Trials().new_trial_docs(
                    tids=[None], specs=[None], results=[None], miscs=[None]
                )
                hyperopt_trial[0] = trial
                hyperopt_trial[0]["tid"] = tid
                hyperopt_trial[0]["misc"]["tid"] = tid
                for key in hyperopt_trial[0]["misc"]["idxs"].keys():
                    hyperopt_trial[0]["misc"]["idxs"][key] = [tid]
                trials1.insert_trial_docs(hyperopt_trial)
                trials1.refresh()
            return trials1

        def proc_train_test(params, X_train, y_train, return_dict):
            return_dict["params"] = copy.deepcopy(params)
            try:
                score, logloss, execution_time = hyperopt_train_test(
                    params, X_train=X_train, y_train=y_train
                )
                return_dict["loss"] = self.best_score - score
                return_dict["time"] = execution_time
                return_dict["log_loss"] = logloss
                return_dict["status"] = hyperopt.STATUS_OK
            except BaseException as e:
                exception_type = f"{type(e).__module__}.{type(e).__name__}"
                try:
                    trainable = create_instance_from_hyperopt_search_space(
                        self.estimator, params
                    )
                    if trainable is None:
                        trial_info = f"hyperparams: {params}"
                    else:
                        trial_info = f'pipeline: """{trainable.pretty_print(show_imports=False)}"""'

                except BaseException:
                    trial_info = f"hyperparams: {params}"
                error_msg = f"Exception caught in Hyperopt: {exception_type}, {traceback.format_exc()}with {trial_info}"
                logger.warning(error_msg + ", setting status to FAIL")
                return_dict["status"] = hyperopt.STATUS_FAIL
                return_dict["error_msg"] = error_msg
                if self.verbose:
                    print(return_dict["error_msg"])

        def get_final_trained_estimator(params, X_train, y_train):
            warnings.filterwarnings("ignore")
            trainable = create_instance_from_hyperopt_search_space(
                self.estimator, params
            )
            if trainable is None:
                return None
            else:
                trained = trainable.fit(X_train, y_train, **fit_params)
                return trained

        def f(params):
            current_time = time.time()
            if (self.max_opt_time is not None) and (
                (current_time - opt_start_time) > self.max_opt_time
            ):
                # if max optimization time set, and we have crossed it, exit optimization completely
                sys.exit(0)
            if self.max_eval_time:
                # Run hyperopt in a subprocess that can be interupted
                manager = multiprocessing.Manager()
                proc_dict: Dict[str, Any] = manager.dict()
                p = multiprocessing.Process(
                    target=proc_train_test, args=(params, X_train, y_train, proc_dict)
                )
                p.start()
                p.join(self.max_eval_time)
                if p.is_alive():
                    p.terminate()
                    p.join()
                    logger.warning(
                        f"Maximum alloted evaluation time exceeded. with hyperparams: {params}, setting status to FAIL"
                    )
                    proc_dict["status"] = hyperopt.STATUS_FAIL
                if "status" not in proc_dict:
                    logger.warning("Corrupted results, setting status to FAIL")
                    proc_dict["status"] = hyperopt.STATUS_FAIL
            else:
                proc_dict = {}
                proc_train_test(params, X_train, y_train, proc_dict)
            return proc_dict

        algo = getattr(hyperopt, self.algo)
        # Search in the search space with defaults
        if self.evals_with_defaults > 0:
            try:
                hyperopt.fmin(
                    f,
                    self.search_space_with_defaults,
                    algo=algo.suggest,
                    max_evals=self.evals_with_defaults,
                    trials=self._default_trials,
                    rstate=np.random.RandomState(SEED),
                    show_progressbar=self.show_progressbar,
                )
            except SystemExit:
                logger.warning(
                    "Maximum alloted optimization time exceeded. Optimization exited prematurely"
                )
            except AllTrialsFailed:
                self._best_estimator = None
                if hyperopt.STATUS_OK not in self._trials.statuses():
                    raise ValueError(
                        "Error from hyperopt, none of the trials succeeded."
                    )

        try:
            hyperopt.fmin(
                f,
                self.search_space,
                algo=algo.suggest,
                max_evals=self.max_evals - self.evals_with_defaults,
                trials=self._trials,
                rstate=np.random.RandomState(SEED),
                show_progressbar=self.show_progressbar,
            )
        except SystemExit:
            logger.warning(
                "Maximum alloted optimization time exceeded. Optimization exited prematurely"
            )
        except AllTrialsFailed:
            self._best_estimator = None
            if hyperopt.STATUS_OK not in self._trials.statuses():
                self._summarize_statuses()
                raise ValueError("Error from hyperopt, none of the trials succeeded.")
        self._trials = merge_trials(self._trials, self._default_trials)
        if self.show_progressbar:
            self._summarize_statuses()
        try:
            best_trial = self._trials.best_trial
            val_loss = self._trials.best_trial["result"]["loss"]
            if len(self._default_trials) > 0:
                default_val_loss = self._default_trials.best_trial["result"]["loss"]
                if default_val_loss < val_loss:
                    best_trial = self._default_trials.best_trial
            best_params = best_trial["result"]["params"]
            logger.info(
                "best score: {:.1%}\nbest hyperparams found using {} hyperopt trials: {}".format(
                    self.best_score - self._trials.average_best_error(),
                    self.max_evals,
                    best_params,
                )
            )
            trained = get_final_trained_estimator(best_params, X_train, y_train)
            self._best_estimator = trained
        except BaseException as e:
            logger.warning(
                "Unable to extract the best parameters from optimization, the error: {}".format(
                    e
                )
            )
            self._best_estimator = None

        return self

    def predict(self, X_eval, **predict_params):
        import warnings

        warnings.filterwarnings("ignore")
        if self._best_estimator is None:
            raise ValueError(
                "Can not predict as the best estimator is None. Either an attempt to call `predict` "
                "before calling `fit` or all the trials during `fit` failed."
            )
        trained = self._best_estimator
        try:
            predictions = trained.predict(X_eval, **predict_params)
        except ValueError as e:
            logger.warning(
                "ValueError in predicting using Hyperopt:{}, the error is:{}".format(
                    trained, e
                )
            )
            predictions = None

        return predictions

    def summary(self):
        """Table summarizing the trial results (ID, loss, time, log_loss, status).

        Returns
        -------
        result : DataFrame"""

        def make_record(trial_dict):
            return {
                "name": f'p{trial_dict["tid"]}',
                "tid": trial_dict["tid"],
                "loss": trial_dict["result"].get("loss", float("nan")),
                "time": trial_dict["result"].get("time", float("nan")),
                "log_loss": trial_dict["result"].get("log_loss", float("nan")),
                "status": trial_dict["result"]["status"],
            }

        records = [make_record(td) for td in self._trials.trials]
        result = pd.DataFrame.from_records(records, index="name")
        return result

    def get_pipeline(self, pipeline_name=None, astype="lale"):
        """Retrieve one of the trials.

        Parameters
        ----------
        pipeline_name : union type, default None

            - string
                Key for table returned by summary(), return a trainable pipeline.

            - None
                When not specified, return the best trained pipeline found.

        astype : 'lale' or 'sklearn', default 'lale'
            Type of resulting pipeline.

        Returns
        -------
        result : Trained operator if best, trainable operator otherwise."""
        best_name = None
        if self._best_estimator is not None:
            best_name = f'p{self._trials.best_trial["tid"]}'
        if pipeline_name is None:
            pipeline_name = best_name
        if pipeline_name == best_name:
            result = getattr(self, "_best_estimator", None)
        else:
            assert pipeline_name is not None
            tid = int(pipeline_name[1:])
            params = self._trials.trials[tid]["result"]["params"]
            result = create_instance_from_hyperopt_search_space(self.estimator, params)
        if result is None or astype == "lale":
            return result
        assert astype == "sklearn", astype
        return result.export_to_sklearn_pipeline()


_hyperparams_schema = {
    "allOf": [
        {
            "type": "object",
            "required": [
                "estimator",
                "max_evals",
                "cv",
                "handle_cv_failure",
                "max_opt_time",
                "pgo",
                "show_progressbar",
            ],
            "relevantToOptimizer": ["estimator", "max_evals", "cv"],
            "additionalProperties": False,
            "properties": {
                "estimator": schema_estimator,
                "scoring": schema_scoring_single,
                "best_score": schema_best_score_single,
                "args_to_scorer": {
                    "anyOf": [
                        {"type": "object"},  # Python dictionary
                        {"enum": [None]},
                    ],
                    "description": "A dictionary of additional keyword arguments to pass to the scorer. Used for cases where the scorer has a signature such as ``scorer(estimator, X, y, **kwargs)``.",
                    "default": None,
                },
                "cv": schema_cv,
                "handle_cv_failure": {
                    "description": """How to deal with cross validation failure for a trial.

If True, continue the trial by doing a 80-20 percent train-validation
split of the dataset input to fit and report the score on the
validation part. If False, terminate the trial with FAIL status.""",
                    "type": "boolean",
                    "default": False,
                },
                "verbose": {
                    "description": """Whether to print errors from each of the trials if any.
This is also logged using logger.warning.""",
                    "type": "boolean",
                    "default": False,
                },
                "show_progressbar": {
                    "description": "Display progress bar during optimization.",
                    "type": "boolean",
                    "default": True,
                },
                "algo": {
                    "description": "Algorithm for searching the space.",
                    "anyOf": [
                        {
                            "enum": ["tpe"],
                            "description": "tree-structured Parzen estimator: https://proceedings.neurips.cc/paper/2011/hash/86e8f7ab32cfd12577bc2619bc635690-Abstract.html",
                        },
                        {"enum": ["atpe"], "description": "adaptive TPE"},
                        {"enum": ["rand"], "description": "random search"},
                        {
                            "enum": ["anneal"],
                            "description": "variant on random search that takes some advantage of a smooth response surface",
                        },
                    ],
                    "default": "tpe",
                },
                "max_evals": {
                    "description": "Number of trials of Hyperopt search.",
                    "type": "integer",
                    "minimum": 1,
                    "default": 50,
                },
                "frac_evals_with_defaults": {
                    "description": """Sometimes, using default values of hyperparameters works quite well.
This value would allow a fraction of the trials to use default values. Hyperopt searches the entire search space
for (1-frac_evals_with_defaults) fraction of max_evals.""",
                    "type": "number",
                    "minimum": 0.0,
                    "default": 0,
                },
                "max_opt_time": schema_max_opt_time,
                "max_eval_time": {
                    "description": "Maximum amout of time in seconds for each evaluation.",
                    "anyOf": [
                        {"type": "number", "minimum": 0.0},
                        {"description": "No runtime bound.", "enum": [None]},
                    ],
                    "default": None,
                },
                "pgo": {
                    "anyOf": [{"description": "lale.search.PGO"}, {"enum": [None]}],
                    "default": None,
                },
            },
        }
    ]
}

_input_fit_schema = {
    "type": "object",
    "required": ["X", "y"],
    "properties": {"X": {}, "y": {}},
}
_input_predict_schema = {"type": "object", "required": ["X"], "properties": {"X": {}}}

_output_predict_schema: Dict[str, Any] = {}

_combined_schemas = {
    "description": """Hyperopt_ is a popular open-source Bayesian optimizer.

.. _Hyperopt: https://github.com/hyperopt/hyperopt

Examples
--------
>>> from lale.lib.sklearn import LogisticRegression as LR
>>> clf = Hyperopt(estimator=LR, cv=3, max_evals=5)
>>> from sklearn import datasets
>>> diabetes = datasets.load_diabetes()
>>> X = diabetes.data[:150]
>>> y = diabetes.target[:150]
>>> trained = clf.fit(X, y)
>>> predictions = trained.predict(X)

Other scoring metrics:

>>> from sklearn.metrics import make_scorer, f1_score
>>> clf = Hyperopt(estimator=LR,
...    scoring=make_scorer(f1_score, average='macro'), cv=3, max_evals=5)
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.lale.hyperopt.html",
    "import_from": "lale.lib.lale",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
    },
}


Hyperopt = lale.operators.make_operator(_HyperoptImpl, _combined_schemas)

lale.docstrings.set_docstrings(Hyperopt)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import warnings

from lale.lib.lale import HalvingGridSearchCV
from lale.lib.sklearn import (
    PCA,
    KNeighborsClassifier,
    KNeighborsRegressor,
    LinearRegression,
    LogisticRegression,
    MinMaxScaler,
    Normalizer,
    StandardScaler,
)


class TestAutoConfigureClassification(unittest.TestCase):
    def setUp(self):
        from sklearn.datasets import load_iris
        from sklearn.model_selection import train_test_split

        data = load_iris()
        X, y = data.data, data.target
        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(X, y)

    def test_with_halving_gridsearchcv(self):
        from lale.lib.lale import HalvingGridSearchCV, NoOp
        from lale.lib.sklearn import PCA, LogisticRegression

        warnings.simplefilter("ignore")
        planned_pipeline = (PCA | NoOp) >> LogisticRegression
        best_pipeline = planned_pipeline.auto_configure(
            self.X_train,
            self.y_train,
            optimizer=HalvingGridSearchCV,
            cv=3,
            scoring="accuracy",
            lale_num_samples=1,
            lale_num_grids=1,
        )
        _ = best_pipeline.predict(self.X_test)
        assert best_pipeline is not None

    def test_runtime_limit_hoc(self):
        import time

        planned_pipeline = (MinMaxScaler | Normalizer) >> (
            LogisticRegression | KNeighborsClassifier
        )
        from sklearn.datasets import load_iris

        X, y = load_iris(return_X_y=True)

        max_opt_time = 10.0
        hoc = HalvingGridSearchCV(
            estimator=planned_pipeline,
            cv=3,
            scoring="accuracy",
            max_opt_time=max_opt_time,
        )
        start = time.time()
        with self.assertRaises(BaseException):
            _ = hoc.fit(X, y)
        end = time.time()
        opt_time = end - start
        rel_diff = (opt_time - max_opt_time) / max_opt_time
        assert (
            rel_diff < 0.7
        ), "Max time: {}, Actual time: {}, relative diff: {}".format(
            max_opt_time, opt_time, rel_diff
        )

    def test_runtime_limit_hor(self):
        import time

        planned_pipeline = (MinMaxScaler | Normalizer) >> LinearRegression
        from sklearn.datasets import load_boston

        X, y = load_boston(return_X_y=True)

        max_opt_time = 3.0
        hor = HalvingGridSearchCV(
            estimator=planned_pipeline,
            cv=3,
            max_opt_time=max_opt_time,
            scoring="r2",
        )
        start = time.time()
        with self.assertRaises(BaseException):
            _ = hor.fit(X[:500, :], y[:500])
        end = time.time()
        opt_time = end - start
        rel_diff = (opt_time - max_opt_time) / max_opt_time
        assert (
            rel_diff < 0.2
        ), "Max time: {}, Actual time: {}, relative diff: {}".format(
            max_opt_time, opt_time, rel_diff
        )


class TestGridSearchCV(unittest.TestCase):
    def test_manual_grid(self):
        from sklearn.datasets import load_iris

        from lale.lib.lale import HalvingGridSearchCV
        from lale.lib.sklearn import SVC

        warnings.simplefilter("ignore")

        from lale import wrap_imported_operators

        wrap_imported_operators()
        iris = load_iris()
        parameters = {"kernel": ("linear", "rbf"), "C": [1, 10]}
        svc = SVC()
        clf = HalvingGridSearchCV(estimator=svc, param_grid=parameters)
        clf.fit(iris.data, iris.target)
        clf.predict(iris.data)

    @unittest.skip("Currently flaky")
    def test_with_halving_gridsearchcv_auto_wrapped_pipe1(self):
        from sklearn.datasets import load_iris
        from sklearn.metrics import accuracy_score, make_scorer

        lr = LogisticRegression()
        pca = PCA()
        trainable = pca >> lr

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            from lale.lib.lale import HalvingGridSearchCV

            clf = HalvingGridSearchCV(
                estimator=trainable,
                lale_num_samples=2,
                lale_num_grids=2,
                cv=2,
                scoring=make_scorer(accuracy_score),
            )
            iris = load_iris()
            clf.fit(iris.data, iris.target)

    @unittest.skip("Currently flaky")
    def test_with_halving_gridsearchcv_auto_wrapped_pipe2(self):
        from sklearn.datasets import load_iris
        from sklearn.metrics import accuracy_score, make_scorer

        lr = LogisticRegression()
        pca1 = PCA()
        pca1._name = "PCA1"
        pca2 = PCA()
        pca2._name = "PCA2"
        trainable = (pca1 | pca2) >> lr

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            from lale.lib.lale import HalvingGridSearchCV

            clf = HalvingGridSearchCV(
                estimator=trainable,
                lale_num_samples=2,
                lale_num_grids=3,
                cv=2,
                scoring=make_scorer(accuracy_score),
            )
            iris = load_iris()
            clf.fit(iris.data, iris.target)


class TestKNeighborsRegressor(unittest.TestCase):
    def setUp(self):
        from sklearn.datasets import load_diabetes
        from sklearn.model_selection import train_test_split

        all_X, all_y = load_diabetes(return_X_y=True)
        # 15 samples, small enough so folds are likely smaller than n_neighbors
        self.train_X, self.test_X, self.train_y, self.test_y = train_test_split(
            all_X, all_y, train_size=15, test_size=None, shuffle=True, random_state=42
        )

    def test_halving_gridsearch(self):
        planned = KNeighborsRegressor
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            trained = planned.auto_configure(
                self.train_X,
                self.train_y,
                optimizer=HalvingGridSearchCV,
                cv=3,
                scoring="r2",
            )
        _ = trained.predict(self.test_X)


class TestStandardScaler(unittest.TestCase):
    def setUp(self):
        import scipy.sparse
        from sklearn.datasets import load_iris
        from sklearn.model_selection import train_test_split

        # from lale.datasets.data_schemas import add_schema
        all_X, all_y = load_iris(return_X_y=True)
        denseTrainX, self.test_X, self.train_y, self.test_y = train_test_split(
            all_X, all_y, train_size=0.8, test_size=0.2, shuffle=True, random_state=42
        )
        # self.train_X = add_schema(scipy.sparse.csr_matrix(denseTrainX))
        self.train_X = scipy.sparse.csr_matrix(denseTrainX)

    def test_halving_gridsearch(self):
        planned = StandardScaler >> LogisticRegression().freeze_trainable()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            trained = planned.auto_configure(
                self.train_X,
                self.train_y,
                optimizer=HalvingGridSearchCV,
                cv=3,
                scoring="r2",
            )
        _ = trained.predict(self.test_X)



import unittest

import lale.datasets
from lale.grammar import Grammar
from lale.lib.lale import ConcatFeatures as Concat
from lale.lib.lale import Hyperopt, NoOp
from lale.lib.sklearn import PCA
from lale.lib.sklearn import AdaBoostClassifier as Boost
from lale.lib.sklearn import KNeighborsClassifier as KNN
from lale.lib.sklearn import LogisticRegression as LR
from lale.lib.sklearn import StandardScaler as Scaler
from lale.operators import PlannedOperator, PlannedPipeline, TrainedOperator


class TestGrammar(unittest.TestCase):
    def setUp(self):
        (
            (self.train_X, self.train_y),
            (self.test_X, self.test_y),
        ) = lale.datasets.load_iris_df()

    def test_grammar_simple(self):
        g = Grammar()
        g.start = g.estimator
        g.estimator = (NoOp | g.transformer) >> g.prim_est
        g.transformer = (NoOp | g.transformer) >> g.prim_tfm

        g.prim_est = LR | KNN
        g.prim_tfm = PCA | Scaler

        generated = g.unfold(6)
        sample = g.sample(6)

        # unfold and sample return a PlannedOperator
        assert isinstance(generated, PlannedOperator)
        assert isinstance(sample, PlannedOperator)

        # test getter for methods other than Nonterminal
        if isinstance(generated, PlannedPipeline):
            assert generated._name.startswith("pipeline")

        try:
            gtrainer = Hyperopt(estimator=generated, max_evals=3, scoring="r2")
            gtrained = gtrainer.fit(self.train_X, self.train_y)
            assert isinstance(gtrained.get_pipeline(), TrainedOperator)
        except ValueError:
            # None of the trials succeeded
            pass

        try:
            strainer = Hyperopt(estimator=sample, max_evals=3, scoring="r2")
            strained = strainer.fit(self.train_X, self.train_y)
            assert isinstance(strained.get_pipeline(), TrainedOperator)
        except ValueError:
            # None of the trials succeeded
            pass

    def test_grammar_all_combinator(self):
        g = Grammar()

        g.start = g.estimator
        g.estimator = g.term_est | g.transformer >> g.term_est
        g.term_est = g.prim_est | g.ensemble
        g.ensemble = Boost(base_estimator=LR)
        g.transformer = g.union_tfm | g.union_tfm >> g.transformer
        g.union_tfm = g.prim_tfm | g.union_body >> Concat
        g.union_body = g.transformer | g.transformer & g.union_body

        g.prim_est = LR | KNN
        g.prim_tfm = PCA | Scaler
        g.ensembler = Boost

        generated = g.unfold(7)
        sample = g.sample(7)
        assert isinstance(generated, PlannedOperator)
        assert isinstance(sample, PlannedOperator)

        # Train
        try:
            gtrainer = Hyperopt(estimator=generated, max_evals=3, scoring="r2")
            gtrained = gtrainer.fit(self.train_X, self.train_y)
            assert isinstance(gtrained.get_pipeline(), TrainedOperator)
        except ValueError:
            # None of the trials succeeded
            pass

        try:
            strainer = Hyperopt(estimator=sample, max_evals=3, scoring="r2")
            strained = strainer.fit(self.train_X, self.train_y)
            assert isinstance(strained.get_pipeline(), TrainedOperator)
        except ValueError:
            # None of the trials succeeded
            pass



# Copyright 2020 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from autoai_ts_libs.srom.estimators.time_series.models.srom_estimators import (  # type: ignore # noqa
    DifferenceFlattenAutoEnsembler as model_to_be_wrapped,
)

import lale.docstrings
import lale.operators


class _DifferenceFlattenAutoEnsemblerImpl:
    def __init__(
        self,
        feature_columns,
        target_columns,
        lookback_win,
        pred_win,
        time_column=-1,
        init_time_optimization=False,
        data_transformation_scheme=None,
        total_execution_time=2,
        execution_time_per_pipeline=1,
        dag_granularity="default",
        execution_platform="spark_node_random_search",
        n_leaders_for_ensemble=5,
        n_estimators_for_pred_interval=30,
        max_samples_for_pred_interval=1.0,
        multistep_prediction_strategy=None,
        multistep_prediction_win=1,
        ensemble_type="voting",
        multivariate_column_encoding=False,
        store_lookback_history=False,
        n_jobs=-1,
        estimator=None,
    ):
        self._hyperparams = {
            "feature_columns": feature_columns,
            "target_columns": target_columns,
            "lookback_win": lookback_win,
            "pred_win": pred_win,
            "time_column": time_column,
            "init_time_optimization": init_time_optimization,
            "data_transformation_scheme": data_transformation_scheme,
            "total_execution_time": total_execution_time,
            "execution_time_per_pipeline": execution_time_per_pipeline,
            "dag_granularity": dag_granularity,
            "execution_platform": execution_platform,
            "n_leaders_for_ensemble": n_leaders_for_ensemble,
            "n_estimators_for_pred_interval": n_estimators_for_pred_interval,
            "max_samples_for_pred_interval": max_samples_for_pred_interval,
            "multistep_prediction_strategy": multistep_prediction_strategy,
            "multistep_prediction_win": multistep_prediction_win,
            "ensemble_type": ensemble_type,
            "multivariate_column_encoding": multivariate_column_encoding,
            "store_lookback_history": store_lookback_history,
            "n_jobs": n_jobs,
            "estimator": estimator,
        }
        self._wrapped_model = model_to_be_wrapped(**self._hyperparams)

    def fit(self, X, y):
        self._wrapped_model.fit(X, y)
        return self

    def predict(self, X=None, **predict_params):
        return self._wrapped_model.predict(X, **predict_params)

    def predict_proba(self, X=None):
        return self._wrapped_model.predict_proba(X)


_hyperparams_schema = {
    "allOf": [
        {
            "description": "This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters.",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "feature_columns",
                "target_columns",
                "lookback_win",
                "pred_win",
                "time_column",
                "init_time_optimization",
                "data_transformation_scheme",
                "total_execution_time",
                "execution_time_per_pipeline",
                "dag_granularity",
                "execution_platform",
                "n_leaders_for_ensemble",
                "n_estimators_for_pred_interval",
                "max_samples_for_pred_interval",
                "multistep_prediction_strategy",
                "multistep_prediction_win",
                "ensemble_type",
                "multivariate_column_encoding",
                "store_lookback_history",
                "n_jobs",
                "estimator",
            ],
            "relevantToOptimizer": [],
            "properties": {
                "feature_columns": {
                    "description": """""",
                    "type": "array",
                    "items": {"type": "integer", "min": 0},  # Positive integers.
                    # "default": None, #TODO: This doesn't have a default value. is that ok?
                },
                "target_columns": {
                    "description": """Column indices for columns to be forecasted.""",
                    "type": "array",
                    "items": {"type": "integer", "min": 0},  # Positive integers.
                    # "default": None, #TODO: This doesn't have a default value. is that ok?
                },
                "lookback_win": {"description": "", "type": "integer"},
                "pred_win": {"description": "", "type": "integer"},
                "time_column": {
                    "description": "",
                    "type": "integer",
                    "default": -1,
                },
                "init_time_optimization": {
                    "description": "",
                    "type": "boolean",
                    "default": False,
                },
                "data_transformation_scheme": {
                    "description": "",
                    "enum": [
                        "auto",
                        "log",
                        "mean_division",
                        "mean_substraction",
                        "mean_division_log",
                        "mean_substraction_log",
                        "sqrt",
                        "reciprocal",
                        "anscombe",
                        "fisher",
                        None,
                    ],
                    "default": None,
                },
                "total_execution_time": {
                    "description": "",
                    "type": "integer",
                    "default": 2,
                },
                "execution_time_per_pipeline": {
                    "description": "",
                    "type": "integer",
                    "default": 1,
                },
                "dag_granularity": {
                    "description": "",
                    "enum": [
                        "default",
                        "tiny",
                        "flat",
                        "multioutput_tiny",
                        "multioutput_flat",
                        "regression_chain_flat",
                        "regression_chain_tiny",
                        "MIMO_flat",
                        "MINO_complete_flat",
                        "multi_xgboost",
                        "multi_lgbm",
                        "xgboost",
                        "lgbm",
                    ],
                    "default": "default",
                },
                "execution_platform": {
                    "description": "",
                    "enum": [
                        "spark_node_random_search"
                    ],  # TODO: are there any other values?
                    "default": "spark_node_random_search",
                },
                "n_leaders_for_ensemble": {
                    "description": "",
                    "type": "integer",
                    "default": 5,
                },
                "n_estimators_for_pred_interval": {
                    "description": "",
                    "type": "integer",
                    "default": 30,
                },
                "max_samples_for_pred_interval": {
                    "description": "",
                    "type": "number",
                    "default": 1.0,
                },
                "multistep_prediction_strategy": {
                    "description": "",
                    "enum": [None, "recursive", "multioutput"],
                    "default": None,
                },
                "multistep_prediction_win": {
                    "description": "",
                    "type": "integer",
                    "default": 1,
                },
                "ensemble_type": {
                    "description": "",
                    "enum": ["voting"],
                    "default": "voting",
                },
                "multivariate_column_encoding": {
                    "description": "",
                    "type": "boolean",
                    "default": False,
                },
                "store_lookback_history": {
                    "description": "",
                    "type": "boolean",
                    "default": False,
                },
                "n_jobs": {
                    "description": "",
                    "anyOf": [
                        {"description": "Use all processors.", "enum": [-1]},
                        {
                            "description": "Number of CPU cores.",
                            "type": "integer",
                            "minimum": 1,
                        },
                    ],
                    "default": -1,
                },
                "estimator": {
                    "description": "estimator object",
                    "anyOf": [{"laleType": "Any"}, {"enum": [None]}],
                    "default": None,
                },
            },
        }
    ]
}

_input_fit_schema = {
    "type": "object",
    "required": ["X", "y"],
    "additionalProperties": False,
    "properties": {
        "X": {  # Handles 1-D arrays as well
            "anyOf": [
                {"type": "array", "items": {"laleType": "Any"}},
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"laleType": "Any"}},
                },
            ]
        },
        "y": {"laleType": "Any"},
    },
}

_input_predict_schema = {
    "type": "object",
    "additionalProperties": False,
    "required": ["X"],
    "properties": {
        "X": {  # Handles 1-D arrays as well
            "anyOf": [
                {"enum": [None]},
                {"type": "array", "items": {"laleType": "Any"}},
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"laleType": "Any"}},
                },
            ]
        }
    },
}

_output_predict_schema = {
    "description": "Features; the outer array is over samples.",
    "anyOf": [
        {"type": "array", "items": {"laleType": "Any"}},
        {"type": "array", "items": {"type": "array", "items": {"laleType": "Any"}}},
    ],
}

_input_predict_proba_schema = {
    "type": "object",
    "additionalProperties": False,
    "required": ["X"],
    "properties": {
        "X": {  # Handles 1-D arrays as well
            "anyOf": [
                {"enum": [None]},
                {"type": "array", "items": {"laleType": "Any"}},
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"laleType": "Any"}},
                },
            ]
        }
    },
}

_output_predict_proba_schema = {
    "description": "Features; the outer array is over samples.",
    "anyOf": [
        {"type": "array", "items": {"laleType": "Any"}},
        {"type": "array", "items": {"type": "array", "items": {"laleType": "Any"}}},
    ],
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """Operator from `autoai_ts_libs`_.

.. _`autoai_ts_libs`: https://pypi.org/project/autoai-ts-libs""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.autoai_ts_libs.difference_flatten_auto_ensembler.html",
    "import_from": "autoai_ts_libs.srom.estimators.time_series.models.srom_estimators",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "forecaster"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
        "input_predict_proba": _input_predict_proba_schema,
        "output_predict_proba": _output_predict_proba_schema,
    },
}

DifferenceFlattenAutoEnsembler = lale.operators.make_operator(
    _DifferenceFlattenAutoEnsemblerImpl, _combined_schemas
)

lale.docstrings.set_docstrings(DifferenceFlattenAutoEnsembler)
